from rest_framework.serializers import CharField
from ai_kit_auth.serializers import UserModel, UserSerializer


class CustomUserSerializer(UserSerializer):
    class Meta:
        model = UserModel
        fields = ["id", "email", "username", "is_active"]

    def get_test_string(self):
        return "test"



import unicodedata
import uuid

from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import UserChangeForm, UserCreationForm
from django.forms import EmailField, CharField
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from .services import send_activation_by_admin_mail
from .settings import api_settings

User = get_user_model()


class UsernameField(CharField):
    def to_python(self, value):
        if not value:
            value = str(uuid.uuid4())
        return unicodedata.normalize("NFKC", super().to_python(value))

    def widget_attrs(self, widget):
        return {
            **super().widget_attrs(widget),
            "autocapitalize": "none",
            "autocomplete": "username",
        }


class AIUserCreationForm(UserCreationForm):
    class Meta:
        model = User
        fields = ("email",)

    username = UsernameField(
        label=_("Username"), required=api_settings.USERNAME_REQUIRED
    )
    email = EmailField(label=_("Email address"), required=True)

    def clean_email(self):
        email = self.cleaned_data["email"]
        if User.objects.filter(email=email).exists():
            raise ValidationError(
                _("Another user with this email already exists!"), code="unique_email"
            )
        return email

    def save(self, commit=True):
        instance = super(AIUserCreationForm, self).save(commit=False)
        instance.is_active = False
        if commit:
            instance.save()
            send_activation_by_admin_mail(instance)
        else:
            # replace the save call, so that an email is send after
            # saving the instance
            bound_save_method = instance.save

            def instance_save_hook(*args, **kwargs):
                bound_save_method(*args, **kwargs)
                send_activation_by_admin_mail(instance)

            instance.save = instance_save_hook
        return instance


class AIUserChangeForm(UserChangeForm):
    class Meta:
        model = User
        fields = ("email",)

    email = EmailField(label=_("Email address"), required=True)

    def clean_email(self):
        email = self.cleaned_data["email"]
        if User.objects.exclude(pk=self.instance.pk).filter(email=email).exists():
            raise ValidationError(
                _("Another user with this email already exists!"), code="unique_email"
            )
        return email


class AIUserAdmin(UserAdmin):
    form = AIUserChangeForm
    add_form = AIUserCreationForm
    fieldsets = api_settings.ADMIN_FIELDSETS
    add_fieldsets = api_settings.ADMIN_ADD_FIELDSETS


if admin.site.is_registered(User):
    admin.site.unregister(User)
admin.site.register(User, AIUserAdmin)



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.core.mail import EmailMultiAlternatives
from django.template.loader import get_template

from .settings import api_settings

User = get_user_model()


def scramble_id(seq_id):
    """
    This is called a feistel chipher.
    It creates a pseudo random, but unique identifier from a number.
    It can be used instead of sequential primary keys in api functions
    to prevent exposing backend details.
    It is its own inverse, so converting back is done by calling it again.

    In case of an id other than an unsigned integer in the 32 bit range we
    assume its a uuid or similar and we return it as is.

    further reading: https://wiki.postgresql.org/wiki/Pseudo_encrypt
    """

    try:
        val = int(seq_id) & 0xFFFFFFFF  # restrict to u32
    except ValueError:
        # not an int, return as is
        return seq_id

    if val != int(seq_id):
        # not in the range of 0 .. 2^32, also return as is
        return seq_id

    l1 = (val >> 16) & 0xFFFF
    r1 = val & 0xFFFF
    for _ in range(0, 3):
        l2 = r1
        r2 = (
            l1
            ^ int(round((((1366 * r1 + 150889) % 714025) / 714025.0) * 32767))
            & 0xFFFFFFFF
        )
        l1 = l2
        r1 = r2
    return ((r1 & 0xFFFF) << 16) + l1


def send_email(subject, text, html, to_address):
    from_address = settings.DEFAULT_FROM_EMAIL
    msg = EmailMultiAlternatives(subject, text, from_address, [to_address])
    msg.attach_alternative(html, "text/html")
    msg.send()


def make_url(*args):
    return "/".join(s.strip("/") for s in args)


def send_user_activation_mail(user):
    """
    Sends the initial mail for a nonactive user.
    """
    template_plain = get_template(
        api_settings.EMAIL_TEMPLATES.USER_CREATED.BODY_PLAINTEXT
    )
    template_html = get_template(api_settings.EMAIL_TEMPLATES.USER_CREATED.BODY_HTML)
    subject = get_template(api_settings.EMAIL_TEMPLATES.USER_CREATED.TITLE).render()

    ident = str(scramble_id(user.pk))
    token_gen = PasswordResetTokenGenerator()
    token = token_gen.make_token(user)

    url = make_url(
        api_settings.FRONTEND.URL, api_settings.FRONTEND.ACTIVATION_ROUTE, ident, token
    )

    context = {
        "user": user,
        "url": url,
    }

    send_email(
        subject.replace("\n", " "),
        template_plain.render(context),
        template_html.render(context),
        getattr(user, User.EMAIL_FIELD),
    )
    return ident, token


def send_activation_by_admin_mail(user):
    """
    send mail for the password reset
    """
    template_plain = get_template(
        api_settings.EMAIL_TEMPLATES.SET_PASSWORD.BODY_PLAINTEXT
    )
    template_html = get_template(api_settings.EMAIL_TEMPLATES.SET_PASSWORD.BODY_HTML)
    subject = get_template(api_settings.EMAIL_TEMPLATES.SET_PASSWORD.TITLE).render()

    ident = str(scramble_id(user.pk))
    token_gen = PasswordResetTokenGenerator()
    token = token_gen.make_token(user)

    url = make_url(
        api_settings.FRONTEND.URL, api_settings.FRONTEND.RESET_PW_ROUTE, ident, token
    )

    context = {
        "user": user,
        "url": url,
    }

    send_email(
        subject.replace("\n", " "),
        template_plain.render(context),
        template_html.render(context),
        getattr(user, User.EMAIL_FIELD),
    )
    return ident, token


def send_reset_pw_mail(user):
    """
    send mail for the password reset
    """
    template_plain = get_template(
        api_settings.EMAIL_TEMPLATES.RESET_PASSWORD.BODY_PLAINTEXT
    )
    template_html = get_template(api_settings.EMAIL_TEMPLATES.RESET_PASSWORD.BODY_HTML)
    subject = get_template(api_settings.EMAIL_TEMPLATES.RESET_PASSWORD.TITLE).render()

    ident = str(scramble_id(user.pk))
    token_gen = PasswordResetTokenGenerator()
    token = token_gen.make_token(user)

    url = make_url(
        api_settings.FRONTEND.URL, api_settings.FRONTEND.RESET_PW_ROUTE, ident, token
    )

    context = {
        "user": user,
        "url": url,
    }

    send_email(
        subject.replace("\n", " "),
        template_plain.render(context),
        template_html.render(context),
        getattr(user, User.EMAIL_FIELD),
    )
    return ident, token



SECRET_KEY = "SLIDFHVsdkjlr3lrnnlrvnker"

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.admin",
    "rest_framework",
    "ai_kit_auth",
)

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.SessionAuthentication",
    ],
}

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"
ROOT_URLCONF = "ai_kit_auth.urls"

AI_KIT_AUTH = {
    "FRONTEND": {"URL": "example.com"},
    "USER_SERIALIZER": "ai_kit_auth.tests.serializers.CustomUserSerializer",
}



"""
Settings module for ai_kit_auth.

Settings are namespaces under AI_KIT_AUTH in
the settings module. So your projects settings.py might look like

AI_KIT_AUTH = {
    "ACTIVATION_ROUTE": "aktivierung",
    ...
}

To access the settings, a api_settings object is provided that gives back the
configured settings or falls back on defaults
"""

from django.conf import settings
from django.test.signals import setting_changed
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from importlib import import_module

DEFAULTS = {
    "EMAIL_TEMPLATES": {
        "USER_CREATED": {
            "TITLE": "ai_kit_auth/user_created_title.txt",
            "BODY_PLAINTEXT": "ai_kit_auth/user_created_body.txt",
            "BODY_HTML": "ai_kit_auth/user_created_body.html",
        },
        "RESET_PASSWORD": {
            "TITLE": "ai_kit_auth/reset_password_title.txt",
            "BODY_PLAINTEXT": "ai_kit_auth/reset_password_body.txt",
            "BODY_HTML": "ai_kit_auth/reset_password_body.html",
        },
        "SET_PASSWORD": {
            "TITLE": "ai_kit_auth/reset_password_title.txt",
            "BODY_PLAINTEXT": "ai_kit_auth/reset_password_body.txt",
            "BODY_HTML": "ai_kit_auth/reset_password_body.html",
        },
    },
    "ADMIN_FIELDSETS": (
        (None, {"fields": ("username", "email", "password")}),
        (_("Personal info"), {"fields": ("first_name", "last_name")}),
        (
            _("Permissions"),
            {
                "fields": (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                    "groups",
                    "user_permissions",
                ),
            },
        ),
        (_("Important dates"), {"fields": ("last_login", "date_joined")}),
    ),
    "ADMIN_ADD_FIELDSETS": (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("username", "email", "password1", "password2"),
            },
        ),
    ),
    "USERNAME_REQUIRED": False,
    "USER_SERIALIZER": "ai_kit_auth.serializers.UserSerializer",
    "FRONTEND": {
        "URL": "",
        "ACTIVATION_ROUTE": "/auth/activation/",
        "RESET_PW_ROUTE": "/auth/reset-password/",
    },
}


class UnknownConfigurationError(Exception):
    pass


class InvalidConfigurationError(Exception):
    pass


class APISettings:
    """
    The settings are implemented as an object for two reasons: it allows access
    to the configuration parameters as properties and it enables validation
    of the settings.

    In your code you can use this object like so:

    from ai_kit_auth.settings import api_settings
    print(api_settings.FRONTEND.ACTIVATION_ROUTE)
    """

    def __init__(self, user_settings, default=DEFAULTS):
        for setting in user_settings:
            if setting not in default:
                raise UnknownConfigurationError(setting)
            elif not isinstance(user_settings[setting], type(default[setting])):
                raise InvalidConfigurationError(
                    f"Type of setting {setting} should be {type(default[setting])}, "
                    f"but was {type(user_settings[setting])}."
                )

        merged = {**default, **user_settings}
        self._settings = {
            k: APISettings.convert_settings_node(v, default[k])
            for k, v in merged.items()
        }

    def __getattr__(self, attr):
        if attr == "USER_SERIALIZER" and type(self._settings[attr]) == str:
            module, name = self._settings[attr].rsplit(".", 1)
            self._settings[attr] = getattr(import_module(module), name)
        return self._settings[attr]

    @classmethod
    def convert_settings_node(cls, node, default):
        if isinstance(node, dict):
            return cls(node, default)
        return node


api_settings = APISettings(getattr(settings, "AI_KIT_AUTH", {}))


@receiver(setting_changed)
def reload_api_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "AI_KIT_AUTH":
        global api_settings
        api_settings = APISettings(getattr(settings, "AI_KIT_AUTH", {}))

        # specialized validation
        if not api_settings.FRONTEND.URL:
            raise InvalidConfigurationError(
                "Please configure FRONTEND_URL in the AI_KIT_AUTH namespace"
            )



from django.dispatch import Signal

user_pre_login = Signal(providing_args=["user"])
user_post_login = Signal(providing_args=["user"])
user_pre_logout = Signal(providing_args=["user"])
user_post_logout = Signal(providing_args=["user"])
user_pre_registered = Signal(providing_args=["user_data"])
user_post_registered = Signal(providing_args=["user"])
user_pre_activated = Signal(providing_args=["user"])
user_post_activated = Signal(providing_args=["user"])
user_pre_forgot_password = Signal(providing_args=["user"])
user_post_forgot_password = Signal(providing_args=["user"])
user_pre_reset_password = Signal(providing_args=["user"])
user_post_reset_password = Signal(providing_args=["user"])



from unittest.mock import Mock, patch
from django.urls import reverse
from django.contrib.auth import authenticate
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.middleware.csrf import _compare_salted_tokens
from django.contrib.auth import get_user_model
from rest_framework import status
from rest_framework.test import APITestCase
from model_bakery import baker
from ai_kit_auth import services

from ai_kit_auth.signals import (
    user_pre_login,
    user_post_login,
    user_pre_logout,
    user_post_logout,
    user_pre_registered,
    user_post_registered,
    user_pre_activated,
    user_post_activated,
    user_pre_forgot_password,
    user_post_forgot_password,
    user_pre_reset_password,
    user_post_reset_password,
)

PASSWORD = "jafsdfah24agdsfghasdf"
EMAIL = "example@example.com"
UserModel = get_user_model()

login_url = reverse("ai_kit_auth:login")
logout_url = reverse("ai_kit_auth:logout")
me_url = reverse("ai_kit_auth:me")
validate_password_url = reverse("ai_kit_auth:validate_password")
activate_url = reverse("ai_kit_auth:activate")
register_url = reverse("ai_kit_auth:register")
send_pw_reset_email_url = reverse("ai_kit_auth:send_pw_reset_email")
pw_reset_url = reverse("ai_kit_auth:pw_reset")


class AuthTestCase(APITestCase):
    def setUp(self) -> None:
        self.user = baker.make(UserModel, email=EMAIL)
        self.user.set_password(PASSWORD)
        self.user.save()

        self.client.logout()

    def isLoggedIn(self) -> bool:
        user_id = self.client.session.get("_auth_user_id", None)
        return user_id is not None and int(user_id) == self.user.id


class LoginTests(AuthTestCase):
    def test_login_with_username(self):
        self.assertFalse(self.isLoggedIn())
        response = self.client.post(
            login_url,
            {"ident": self.user.username, "password": PASSWORD},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("csrf" in response.data)
        self.assertEqual(response.data["user"]["username"], self.user.username)
        self.assertEqual(response.data["user"]["id"], self.user.id)
        self.assertEqual(response.data["user"]["email"], self.user.email)
        self.assertTrue(self.isLoggedIn())

    def test_login_with_email(self):
        self.assertFalse(self.isLoggedIn())
        response = self.client.post(
            login_url, {"ident": self.user.email, "password": PASSWORD}, format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("csrf" in response.data)
        self.assertEqual(response.data["user"]["username"], self.user.username)
        self.assertEqual(response.data["user"]["id"], self.user.id)
        self.assertEqual(response.data["user"]["email"], self.user.email)
        self.assertTrue(self.isLoggedIn())

    def test_login_wrong_pw(self):
        self.assertFalse(self.isLoggedIn())
        response = self.client.post(
            login_url,
            {"ident": self.user.email, "password": "wrong" + PASSWORD},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertFalse(self.isLoggedIn())

    @patch("ai_kit_auth.views.login")
    def test_login_emits_signals(self, mock_login):
        received = Mock()

        def receiver_pre(sender, user, **kwargs):
            mock_login.assert_not_called()
            received(pre=True)

        def receiver_post(sender, user, **kwargs):
            mock_login.assert_called()
            received(post=True)

        user_pre_login.connect(receiver_pre)
        user_post_login.connect(receiver_post)

        self.client.post(
            login_url,
            {"ident": self.user.username, "password": PASSWORD},
            format="json",
        )

        user_pre_login.disconnect(receiver_pre)
        user_post_login.disconnect(receiver_post)

        received.assert_any_call(pre=True)
        received.assert_any_call(post=True)

    def test_csrf_token_after_login(self):
        response = self.client.post(
            login_url,
            {"ident": self.user.username, "password": PASSWORD},
            format="json",
        )
        self.assertTrue(
            _compare_salted_tokens(
                response.cookies["csrftoken"].value, response.data["csrf"]
            )
        )


class MeViewTests(AuthTestCase):
    def setUp(self) -> None:
        self.user = baker.make(UserModel, email=EMAIL)
        self.user.set_password(PASSWORD)
        self.user.save()

        self.client.logout()

    def test_access_protected(self):
        self.client.login(username=self.user.username, password=PASSWORD)
        response = self.client.get(me_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("csrf" in response.data)
        self.assertEqual(response.data["user"]["email"], self.user.email)
        self.assertEqual(response.data["user"]["username"], self.user.username)

    def test_access_me_not_logged_in(self):
        response = self.client.get(me_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("csrf" in response.data)
        self.assertEqual(response.data["user"], None)

    def test_custom_user_serializer(self):
        self.client.login(username=self.user.username, password=PASSWORD)
        response = self.client.get(me_url)
        self.assertTrue("is_active" in response.data["user"])


class LogoutTests(AuthTestCase):
    def test_logout(self):
        self.client.login(username=self.user.username, password=PASSWORD)
        self.assertTrue(self.isLoggedIn())
        response = self.client.post(logout_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertFalse(self.isLoggedIn())

    @patch("ai_kit_auth.views.logout")
    def test_logout_emits_signals(self, mock_logout):
        self.client.login(username=self.user.username, password=PASSWORD)
        received = Mock()

        def receiver_pre(sender, user, **kwargs):
            mock_logout.assert_not_called()
            received(pre=True)

        def receiver_post(sender, user, **kwargs):
            mock_logout.assert_called()
            received(post=True)

        user_pre_logout.connect(receiver_pre)
        user_post_logout.connect(receiver_post)

        response = self.client.post(logout_url)

        user_pre_logout.disconnect(receiver_pre)
        user_post_logout.disconnect(receiver_post)

        received.assert_any_call(pre=True)
        received.assert_any_call(post=True)

    def test_access_protected_fail(self):
        self.assertFalse(self.isLoggedIn())
        response = self.client.post(logout_url)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class ValidatePasswordTests(AuthTestCase):
    def test_validate_password(self):
        response = self.client.post(
            validate_password_url,
            {
                "password": "longandvalidpassword",
                "username": "username",
                "email": EMAIL,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_validate_password_with_ident(self):
        ident = str(services.scramble_id(self.user.pk))
        response = self.client.post(
            validate_password_url,
            {"password": PASSWORD, "ident": ident,},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_validate_password_fail(self):
        response = self.client.post(
            validate_password_url,
            {"password": "username", "username": "username", "email": EMAIL},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class ActivateEmailTests(AuthTestCase):
    def test_activate_user(self):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        ident, token = services.send_user_activation_mail(user)
        response = self.client.post(
            activate_url, {"ident": ident, "token": token}, format="json"
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # we have to get the user object again to see the updates
        self.assertTrue(UserModel.objects.get(pk=user.id).is_active)

    def test_activate_user_emits_signals(self):
        received = Mock()

        def receiver_pre(sender, user, **kwargs):
            self.assertFalse(user.is_active)
            received(pre=True)

        def receiver_post(sender, user, **kwargs):
            self.assertTrue(user.is_active)
            received(post=True)

        user_pre_activated.connect(receiver_pre)
        user_post_activated.connect(receiver_post)

        user = baker.make(UserModel, is_active=False, email="to@example.com")
        ident, token = services.send_user_activation_mail(user)
        response = self.client.post(
            activate_url, {"ident": ident, "token": token}, format="json"
        )

        user_pre_activated.disconnect(receiver_pre)
        user_post_activated.disconnect(receiver_post)

        received.assert_any_call(pre=True)
        received.assert_any_call(post=True)


class ResetPWTests(AuthTestCase):
    def test_init_password_reset(self):
        response = self.client.post(send_pw_reset_email_url, {"email": self.user.email})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @patch("ai_kit_auth.views.services.send_reset_pw_mail")
    def test_init_password_reset_emits_signals(self, mock_service):
        received = Mock()

        def receiver_pre(sender, user, **kwargs):
            mock_service.assert_not_called()
            received(pre=True)

        def receiver_post(sender, user, **kwargs):
            mock_service.assert_called()
            received(post=True)

        user_pre_forgot_password.connect(receiver_pre)
        user_post_forgot_password.connect(receiver_post)

        response = self.client.post(send_pw_reset_email_url, {"email": self.user.email})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        user_pre_forgot_password.disconnect(receiver_pre)
        user_post_forgot_password.disconnect(receiver_post)

        received.assert_any_call(pre=True)
        received.assert_any_call(post=True)

    def test_init_password_reset_fail(self):
        response = self.client.post(
            send_pw_reset_email_url, {"email": "invalid_email@example.com"}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_reset_password(self):
        new_password = "new_awesome_password"
        # manually create reset session
        ident = str(services.scramble_id(self.user.pk))
        token_gen = PasswordResetTokenGenerator()
        token = token_gen.make_token(self.user)

        response = self.client.post(
            pw_reset_url,
            {"password": new_password, "ident": ident, "token": token},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(
            authenticate(username=self.user.username, password=new_password)
        )

    def test_reset_password_emits_signals(self):
        received = Mock()

        def receiver_pre(sender, user, **kwargs):
            self.assertIsNone(
                authenticate(username=self.user.username, password=new_password)
            )
            received(pre=True)

        def receiver_post(sender, user, **kwargs):
            self.assertIsNotNone(
                authenticate(username=self.user.username, password=new_password)
            )
            received(post=True)

        user_pre_reset_password.connect(receiver_pre)
        user_post_reset_password.connect(receiver_post)

        new_password = "new_awesome_password"
        # manually create reset session
        ident = str(services.scramble_id(self.user.pk))
        token_gen = PasswordResetTokenGenerator()
        token = token_gen.make_token(self.user)

        response = self.client.post(
            pw_reset_url,
            {"password": new_password, "ident": ident, "token": token},
            format="json",
        )
        user_pre_reset_password.disconnect(receiver_pre)
        user_post_reset_password.disconnect(receiver_post)

        received.assert_any_call(pre=True)
        received.assert_any_call(post=True)

    def test_reset_password_invalid_token(self):
        new_password = "new_awesome_password"
        ident = str(services.scramble_id(self.user.pk))
        response = self.client.post(
            pw_reset_url,
            {"password": new_password, "ident": ident, "token": "notavalidtoken"},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)


class RegisterTests(AuthTestCase):
    def test_register_user(self):
        response = self.client.post(
            register_url,
            {
                "username": "testuser",
                "email": "testuser@example.com",
                "password": PASSWORD,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        user = UserModel.objects.get(email="testuser@example.com")
        self.assertEqual(user.username, "testuser")
        self.assertFalse(user.is_active)
        # Make sure that password is not stored in plain text
        self.assertNotEqual(user.password, PASSWORD)

    def test_register_user_emits_signals(self):
        received = Mock()

        def receiver_pre(sender, user_data, **kwargs):
            self.assertFalse(
                UserModel.objects.filter(username=user_data["username"]).exists()
            )
            received(pre=True)

        def receiver_post(sender, user, **kwargs):
            self.assertIsNotNone(user.pk)
            received(post=True)

        user_pre_registered.connect(receiver_pre)
        user_post_registered.connect(receiver_post)

        response = self.client.post(
            register_url,
            {
                "username": "testuser_emit_signals",
                "email": "testuser_emit_signals@example.com",
                "password": PASSWORD,
            },
            format="json",
        )
        user_pre_registered.disconnect(receiver_pre)
        user_post_registered.disconnect(receiver_post)

        received.assert_any_call(pre=True)
        received.assert_any_call(post=True)

    def test_register_user_same_username_fail(self):
        response = self.client.post(
            register_url,
            {
                "username": self.user.username,
                "email": "testuser@example.de",
                "password": PASSWORD,
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data["username"][0]), "username_unique")

    def test_register_user_same_email_fail(self):
        response = self.client.post(
            register_url,
            {"username": "testuser", "email": self.user.email, "password": PASSWORD,},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data["email"][0]), "email_unique")

    def test_register_user_invalid_password_fail(self):
        response = self.client.post(
            register_url,
            {
                "username": "testuser",
                "email": "testuser@example.com",
                "password": "short",
            },
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data["password"][0]), "password_too_short")

    def test_register_blank_username_when_not_required(self):
        response = self.client.post(
            register_url,
            {"username": "", "email": "a@example.com", "password": PASSWORD},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_register_required_email_error(self):
        response = self.client.post(
            register_url,
            {"username": "username123", "password": PASSWORD},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(str(response.data["email"][0]), "required")

    def test_register_blank_email_error(self):
        response = self.client.post(
            register_url,
            {"username": "username123", "email": "", "password": PASSWORD},
            format="json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue("blank" in response.data["email"])

    def test_password_too_short_error(self):
        response = self.client.post(
            register_url,
            {
                "username": "newuser42",
                "email": "lasjdvh@example.com",
                "password": "b3dkjA3",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue("password_too_short" in response.data["password"])



from unittest import TestCase
from unittest.mock import patch

from ai_kit_auth.settings import (
    APISettings,
    UnknownConfigurationError,
    InvalidConfigurationError,
    reload_api_settings,
)


class APISettingsTests(TestCase):
    def test_raises_if_unknown_setting(self):
        user_settings = {"NOT_A_KEY_IN_DEFAULT": "value"}
        with self.assertRaises(UnknownConfigurationError):
            APISettings(user_settings)

    def test_raises_if_improper_setting(self):
        user_settings = {"USERNAME_REQUIRED": "yes"}
        with self.assertRaises(InvalidConfigurationError):
            APISettings(user_settings)

    def test_values_are_overridden(self):
        user_settings = {"USERNAME_REQUIRED": True}
        api_settings = APISettings(user_settings)
        self.assertEqual(
            user_settings["USERNAME_REQUIRED"], api_settings.USERNAME_REQUIRED
        )

    def test_conversion_is_nested(self):
        api_settings = APISettings({})
        self.assertEqual(api_settings.FRONTEND.URL, "")


class ReloadAPISettingsTests(TestCase):
    @patch("ai_kit_auth.settings.settings")
    def test_raises_if_FRONTEND_URL_is_not_set(self, mock_settings):
        mock_settings["AI_KIT_AUTH"] = {}
        with self.assertRaises(InvalidConfigurationError):
            reload_api_settings(setting="AI_KIT_AUTH")



import unicodedata
import uuid
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.password_validation import (
    get_password_validators,
    validate_password,
)
from django.conf import settings
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.utils import IntegrityError
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from .settings import api_settings
from .signals import user_post_registered
from . import services

UserModel = get_user_model()

# common arguments to the serializer fields
FIELD_ARGS = {
    "required": True,
    "error_messages": {"required": "required", "blank": "blank"},
}


class LoginSerializer(serializers.Serializer):
    ident = serializers.CharField(**FIELD_ARGS)
    password = serializers.CharField(style={"input_type": "password"}, **FIELD_ARGS,)

    def validate(self, attrs):
        ident = attrs.get("ident")
        password = attrs.get("password")
        # if the ident is an email, we have to map it to a username
        try:
            ident = UserModel.objects.get(email__iexact=ident).get_username()
        except UserModel.DoesNotExist:
            pass

        user = authenticate(self.context["request"], username=ident, password=password)

        if not user:
            raise ValidationError("invalid_credentials")

        attrs["user"] = user
        return attrs


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserModel
        fields = ["id", "email", "username"]


class ValidatePasswordSerializer(serializers.Serializer):
    # either ident or email and username (only if configured) is required,
    # but we test that manually
    ident = serializers.CharField(required=False)
    username = serializers.CharField(required=False, allow_blank=True)
    email = serializers.EmailField(required=False, allow_blank=True)
    password = serializers.CharField(
        required=True, error_messages={"required": "required", "blank": "blank"},
    )

    def validate(self, attrs):
        ident = attrs.get("ident")
        username = attrs.get("username")
        email = attrs.get("email")
        password = attrs["password"]

        if ident:
            # we dont need username and/or email, the usermodel should already
            # exist
            try:
                pk = services.scramble_id(ident)
                user = UserModel.objects.get(pk=pk)
            except UserModel.DoesNotExist:
                # if anything goes wrong, we error out
                raise ValidationError("unknown_user")
        else:
            # usermodel does not already exist, we create a one off only for the
            # validation
            user = UserModel(username=username, email=email,)

        try:
            validators = get_password_validators(settings.AUTH_PASSWORD_VALIDATORS)
        except:
            return attrs
        try:
            validate_password(
                password=password, user=user, password_validators=validators,
            )
        except DjangoValidationError as e:
            # convert to error codes since translations are implemented in the
            # frontend
            raise ValidationError({"password": [error.code for error in e.error_list]})
        return attrs


class InitiatePasswordResetSerializer(serializers.Serializer):
    email = serializers.EmailField(**FIELD_ARGS)


class PasswordResetSerializer(serializers.Serializer):
    ident = serializers.CharField(**FIELD_ARGS)
    token = serializers.CharField(**FIELD_ARGS)
    password = serializers.CharField(**FIELD_ARGS)


class RegistrationSerializer(serializers.Serializer):
    username = serializers.CharField(
        required=api_settings.USERNAME_REQUIRED,
        allow_blank=not api_settings.USERNAME_REQUIRED,
        error_messages=FIELD_ARGS["error_messages"],
    )
    password = serializers.CharField(**FIELD_ARGS)
    email = serializers.EmailField(**FIELD_ARGS)

    def validate(self, attrs):
        def normalize(value):
            if not value:
                value = str(uuid.uuid4())
            return unicodedata.normalize("NFKC", value)

        username = normalize(attrs.get("username"))
        email = attrs["email"]
        password = attrs["password"]

        password_serializer = ValidatePasswordSerializer(
            data={"username": username, "email": email, "password": password}
        )
        try:
            password_serializer.is_valid(raise_exception=True)
        except DjangoValidationError as e:
            # convert to error codes since translations are implemented in the
            # frontend
            raise ValidationError({"password": [error.code for error in e.error_list]})

        # make sure email is unique
        if UserModel.objects.filter(email=email).exists():
            raise ValidationError({"email": ["email_unique"]})

        try:
            user = UserModel(username=username, email=email, is_active=False)
            user.set_password(password)
            user.save()
            user_post_registered.send(sender=RegistrationSerializer, user=user)
        except IntegrityError as e:
            raise ValidationError({"username": ["username_unique"]})
        services.send_user_activation_mail(user)
        return attrs



from django.contrib.auth import login, logout, get_user_model, tokens
from rest_framework import status, generics, views
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated, AllowAny
from . import serializers, services
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.middleware import csrf

from .settings import api_settings

from .signals import (
    user_pre_login,
    user_post_login,
    user_pre_logout,
    user_post_logout,
    user_pre_registered,
    user_pre_activated,
    user_post_activated,
    user_pre_forgot_password,
    user_post_forgot_password,
    user_pre_reset_password,
    user_post_reset_password,
)

UserModel = get_user_model()


class LoginView(generics.GenericAPIView):
    # uses standard user model
    serializer_class = serializers.LoginSerializer
    permission_classes = (AllowAny,)
    user_serializer = api_settings.USER_SERIALIZER

    @method_decorator(csrf_protect)
    def post(self, request, *args, **kwargs):
        serializer = self.serializer_class(
            data=self.request.data, context={"request": request}
        )
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data["user"]

        user_pre_login.send(sender=LoginView, user=user)

        login(request, user)

        user_serializer = self.user_serializer(
            instance=user, context={"request": request}
        )

        # the position of this statement is important since the csrf token
        # is rotated on login
        csrf_token = csrf.get_token(request)

        user_post_login.send(sender=LoginView, user=user)

        return Response(
            {"user": user_serializer.data, "csrf": csrf_token,},
            status=status.HTTP_200_OK,
        )


class LogoutView(views.APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            user_pre_logout.send(sender=LogoutView, user=request.user)
            logout(request)
            user_post_logout.send(sender=LogoutView, user=request.user)
        csrf_token = csrf.get_token(request)
        return Response({"csrf": csrf_token}, status=status.HTTP_200_OK)


class MeView(generics.GenericAPIView):
    """
    Barebones user model detail view
    """

    permission_classes = (AllowAny,)
    user_serializer = api_settings.USER_SERIALIZER

    def get(self, request, *args, **kwargs):
        csrf_token = csrf.get_token(request)

        if request.user.is_anonymous:
            user_data = None
        else:
            user_serializer = self.user_serializer(
                instance=request.user, context={"request": request}
            )
            user_data = user_serializer.data
        return Response(
            {"user": user_data, "csrf": csrf_token,}, status=status.HTTP_200_OK
        )


class RegistrationView(generics.GenericAPIView):

    permission_classes = (AllowAny,)

    serializer_class = serializers.RegistrationSerializer

    def post(self, request, *args, **kwargs):
        user_pre_registered.send(sender=RegistrationView, user_data=request.data)
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        return Response({}, status=status.HTTP_201_CREATED)


class ValidatePassword(views.APIView):
    """
    Endpoint to validate the password without trying to register an account.
    Can be used to show the user error messages on the fly
    """

    serializer_class = serializers.ValidatePasswordSerializer
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        return Response({}, status=status.HTTP_200_OK)


class ActivateUser(views.APIView):
    """
    Endpoint for double opt in user activation
    """

    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        try:
            ident = request.data["ident"]
            token = request.data["token"]
            pk = services.scramble_id(ident)
            user = UserModel.objects.get(pk=pk)
            assert tokens.PasswordResetTokenGenerator().check_token(user, token)
        except (
            TypeError,
            ValueError,
            OverflowError,
            AssertionError,
            UserModel.DoesNotExist,
        ):
            return Response(
                {"error": "activation_link_invalid"}, status=status.HTTP_400_BAD_REQUEST
            )
        user_pre_activated.send(sender=ActivateUser, user=user)
        user.is_active = True
        user.save()
        user_post_activated.send(sender=ActivateUser, user=user)
        return Response(status=status.HTTP_200_OK)


class InitiatePasswordResetView(views.APIView):
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        try:
            user = UserModel.objects.get(email__iexact=request.data["email"])
            user_pre_forgot_password.send(sender=InitiatePasswordResetView, user=user)
            services.send_reset_pw_mail(user)
            user_post_forgot_password.send(sender=InitiatePasswordResetView, user=user)
        except UserModel.DoesNotExist:
            pass
        # always return OK
        return Response(status=status.HTTP_200_OK)


class ResetPassword(views.APIView):
    """
    Endpoint to reset the password for a user. It also activates the user if
    the active flag is not already set
    """

    serializer_class = serializers.PasswordResetSerializer
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        ident = request.data["ident"]
        token = request.data["token"]
        password = request.data["password"]

        try:
            # we need a valid session to reset the password
            pk = services.scramble_id(ident)
            user = UserModel.objects.get(pk=pk)
            assert tokens.PasswordResetTokenGenerator().check_token(user, token)
        except (
            TypeError,
            ValueError,
            OverflowError,
            AssertionError,
            UserModel.DoesNotExist,
        ):
            # if anything goes wrong, we error out
            return Response(
                {"error": "reset_password_link_invalid"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        user_pre_reset_password.send(sender=ResetPassword, user=user)
        # reuse the password validation
        serializer = serializers.ValidatePasswordSerializer(
            data={"ident": ident, "password": password,}
        )
        serializer.is_valid(raise_exception=True)

        # ok, everything is fine, we do the actual password reset
        user.is_active = True
        user.set_password(password)
        user.save()
        user_post_reset_password.send(sender=ResetPassword, user=user)
        return Response({}, status=status.HTTP_200_OK)



import re
from setuptools import setup

with open("ai_kit_auth/__init__.py", "r") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

setup(
    version=version,
    install_requires=["Django>=3.0.0"],
    tests_require=[],
    test_suite="runtests.runtests",
)



from django.urls import include, path
from . import views

auth_patterns = (
    [
        path(r"login/", views.LoginView.as_view(), name="login"),
        path(r"activate_email/", views.ActivateUser.as_view(), name="activate",),
        path(r"logout/", views.LogoutView.as_view(), name="logout"),
        path(r"me/", views.MeView.as_view(), name="me"),
        path(
            r"validate_password/",
            views.ValidatePassword.as_view(),
            name="validate_password",
        ),
        path(
            r"send_pw_reset_email/",
            views.InitiatePasswordResetView.as_view(),
            name="send_pw_reset_email",
        ),
        path(r"reset_password/", views.ResetPassword.as_view(), name="pw_reset",),
        path(r"register/", views.RegistrationView.as_view(), name="register"),
    ],
    "ai_kit_auth",
)

urlpatterns = [path("", include(auth_patterns))]



from unittest.mock import patch
import uuid
from django.test import TestCase
from django.contrib.auth import get_user_model
from django.core import mail
from model_bakery import baker
from ai_kit_auth.services import (
    scramble_id,
    send_user_activation_mail,
    send_reset_pw_mail,
    send_email,
    make_url,
)


UserModel = get_user_model()


class FeistelChipherTest(TestCase):
    def test_invertable(self):
        for i in list(range(1000)) + [4020, 19432, 599299398]:
            encoded = scramble_id(i)
            self.assertNotEqual(i, encoded)
            self.assertEqual(i, scramble_id(encoded))

    def test_int_as_string(self):
        for i in list(range(1000)) + [4020, 19432, 599299398]:
            encoded = scramble_id(str(i))
            self.assertNotEqual(i, encoded)
            self.assertEqual(i, scramble_id(encoded))

    def test_identity_if_not_int(self):
        ident = "test"
        self.assertEqual(ident, scramble_id(ident))

    def test_out_of_range(self):
        for i in [-1, 0x100000000, 0x29943929990, "string", uuid.uuid4()]:
            self.assertEqual(i, scramble_id(i))


class MailTests(TestCase):
    def test_send_email(self):
        subject = "subject"
        text = "plain text"
        html = "<h1>HTML Version</h1>"
        to_address = "recipient@example.com"
        send_email(subject, text, html, to_address)
        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertEqual(email.subject, subject)
        self.assertEqual(email.body, text)
        self.assertEqual(email.alternatives[0][0], html)
        self.assertEqual(email.to, [to_address])


class MakeURLTests(TestCase):
    def test_concatenates_arguments(self):
        args = [str(i) for i in range(10)]
        self.assertEqual(make_url(*args), "/".join(args))

    def test_strips_slashes(self):
        args = [f"/{i}/" for i in range(10)]
        self.assertEqual(make_url(*args), "/".join(str(i) for i in range(10)))


class ActivationTest(TestCase):
    @patch("ai_kit_auth.services.send_email")
    def test_send_activation_mail(self, mock_send_mail):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        send_user_activation_mail(user)
        mock_send_mail.assert_called()

    def test_user_id_is_scrambled(self):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        ident, token = send_user_activation_mail(user)
        self.assertEqual(ident, str(scramble_id(user.pk)))

    @patch("ai_kit_auth.services.make_url")
    def test_activation_link_is_in_email(self, mock_make_url):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        mock_make_url.return_value = "url to the frontend activation link thingy"
        send_user_activation_mail(user)
        self.assertTrue(mock_make_url.return_value in mail.outbox[0].body)


class InitResetPasswordTest(TestCase):
    @patch("ai_kit_auth.services.send_email")
    def test_send_activation_mail(self, mock_send_mail):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        send_reset_pw_mail(user)
        mock_send_mail.assert_called()

    def test_user_id_is_scrambled(self):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        ident, token = send_reset_pw_mail(user)
        self.assertEqual(ident, str(scramble_id(user.pk)))

    @patch("ai_kit_auth.services.make_url")
    def test_reset_link_is_in_email(self, mock_make_url):
        user = baker.make(UserModel, is_active=False, email="to@example.com")
        mock_make_url.return_value = "url to the frontend reset link thingy"
        send_reset_pw_mail(user)
        self.assertTrue(mock_make_url.return_value in mail.outbox[0].body)



from unittest import TestCase
from unittest.mock import Mock, patch

from django.core.exceptions import ValidationError
from model_bakery import baker

from ai_kit_auth.admin import User, AIUserCreationForm, AIUserChangeForm

EMAIL = "a@example.com"


@patch("ai_kit_auth.admin.UserCreationForm.save", Mock())
class CreateFormTests(TestCase):
    def test_clean_raises_if_email_exists(self):
        baker.make(User, email=EMAIL)

        form = Mock()
        form.cleaned_data = {"email": EMAIL}

        with self.assertRaises(ValidationError):
            AIUserCreationForm.clean_email(form)

    def test_clean_returns_email(self):
        User.objects.filter(email=EMAIL).delete()
        form = Mock()
        form.cleaned_data = {"email": EMAIL}
        self.assertEqual(AIUserCreationForm.clean_email(form), EMAIL)

    @patch("ai_kit_auth.admin.send_activation_by_admin_mail", Mock())
    def test_save_sets_inactive(self):
        instance = AIUserCreationForm().save()
        self.assertFalse(instance.is_active)

    @patch("ai_kit_auth.admin.send_activation_by_admin_mail")
    def test_creation_calls_mail_send_service(self, mock_send_mail):
        instance = AIUserCreationForm().save()
        mock_send_mail.assert_called_with(instance)

    @patch("ai_kit_auth.admin.send_activation_by_admin_mail")
    def test_creation_send_mail_with_saved_user(self, mock_send_mail):
        def check_if_instance_is_saved(ins):
            ins.save.assert_called()

        mock_send_mail.side_effect = check_if_instance_is_saved
        instance = AIUserCreationForm().save()
        mock_send_mail.assert_called_with(instance)

    @patch("ai_kit_auth.admin.send_activation_by_admin_mail")
    def test_save_with_commit_false(self, mock_send_mail):
        instance = AIUserCreationForm().save(commit=False)
        mock_send_mail.assert_not_called()
        instance.save()
        mock_send_mail.assert_called_with(instance)


class ChangeFormTests(TestCase):
    def test_clean_raises_if_email_changes_to_already_taken_one(self):
        baker.make(User, email=EMAIL)
        user = baker.make(User)

        form = AIUserChangeForm(instance=user)
        form.cleaned_data = {"email": EMAIL}

        with self.assertRaises(ValidationError):
            form.clean_email()

    def test_clean_is_okay_if_email_is_the_one_of_instance_and_returns_email(self):
        user = baker.make(User, email=EMAIL)
        form = AIUserChangeForm(instance=user)
        form.cleaned_data = {"email": EMAIL}
        # Should not raise
        self.assertEqual(form.clean_email(), EMAIL)
